# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the `Executor` class, which manages the execution of `Job`s.
This is built upon Python's `multiprocessing` module."""

import multiprocessing
from multiprocessing.context import SpawnProcess
from queue import Empty
from threading import Thread
from typing import Any, Callable, Optional, TYPE_CHECKING, ValuesView
from datetime import datetime
from collections import defaultdict
from dataclasses import asdict, dataclass
import random
import logging
import contextlib
from .job import Job
from . import events

if TYPE_CHECKING:
    from .scheduler import Scheduler

random.seed()
logger = logging.getLogger(__name__)

def _require_running(func: Callable[..., Any]) -> Any:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        executor: Executor = args[0]

        if not executor.is_running():
            raise ExecutorNotSetupException()

        return func(*args, **kwargs)

    return wrapper


@dataclass
class Instance:
    """Class for correlating a `Job` and its process information.

    Attributes:
        job (Job): A `Job` instance.
        id (int): A unique identifier.
        start_time (datetime): When this `Job` instance started.
        process (multiprocessing.SpawnProcess): The associated process with this instance.
    """
    job: Job
    id: int
    start_time: datetime
    process: SpawnProcess


class Executor:
    """Manages the execution of `Job`s by utilizing multiprocessing.

    Attributes:
        _pool_size (int): The number of `Job` processes executing. NOTE: Currently unutilitzed.
        _scheduler (Scheduler): The `Scheduler` which manages this `Executor`.
        _instances (Dict[int, Instance]): Mapping of process instance ids to `Job` process information.
            Instance ids are a way to associate a running `Job` with its process and start time.
        _event_queue (multiprocessing.Queue): Queue for communicating events across processes.
        _event_thread (threading.Thread): Thread that handles the `_event_queue`.
        _running (bool): Tracks the current running state.
        _lock (threading.RLock): A Reentrant lock to handle access to process instances.
        mp_ctx (multiprocessing.SpawnContext): The multiprocessing context to use, namely `Spawn`.
    """

    def __init__(self, pool_size: int, scheduler: 'Scheduler') -> None:
        """Initializes an `Executor`.

        Args:
            pool_size (int): The number of `Job` processes executing. NOTE: Currently unutilitzed.
            scheduler (Scheduler): The `Scheduler` which manages this `Executor`.
        """
        self._pool_size = pool_size  # TODO: use _pool_size
        self._scheduler = scheduler

        self._instances: dict[int, Instance] = {}
        self._jobs_to_instances: dict[str, list[int]] = defaultdict(list)

        self.mp_ctx = multiprocessing.get_context('spawn')

        if TYPE_CHECKING:
            # pylint: disable=unsubscriptable-object
            self._event_queue: Optional[multiprocessing.Queue[events.JobExecutionEvent]] = None
        else:
            self._event_queue: Optional[multiprocessing.Queue] = None

        self._event_thread: Optional[Thread] = None
        self._running = False

        self._lock = self.mp_ctx.RLock()

    def setup(self) -> None:
        """Performs basic setup functions, also starts watching the `_event_queue`.

        The `_event_thread` is ran as a daemon thread.
        """
        if self._running:
            return

        self._event_queue = self.mp_ctx.Queue()

        self._event_thread = Thread(target=self._event_read_loop, daemon=True)
        self._running = True
        self._event_thread.start()

    @_require_running
    def teardown(self, wait: bool = True) -> None:
        """Shuts down `Job` processes and cleans up the `_event_queue`.

        Args:
            wait (bool, optional): Determines if it should wait for processes to finish shutting down.
                Defaults to True.
        """
        logger.debug('Tearing down...')

        processes = [instance.process for instance in self._instances.values()]
        logger.debug(f'Processes: {processes}')

        def get_alive_procs() -> list[SpawnProcess]:
            procs = []
            for proc in processes:
                try:
                    if proc.is_alive():
                        procs.append(proc)
                except ValueError:
                    pass
            return procs

        while alive_processes := get_alive_procs():
            logger.debug(f'Alive processes: {alive_processes}')
            for process in alive_processes:
                logger.debug(f'Process: {process}')
                if not wait:
                    logger.debug('    terminating')
                    process.terminate()
                logger.debug('    joining')
                process.join(1)
                if not process.is_alive():
                    logger.debug('    closing')
                    try:
                        process.close()
                    except ValueError:
                        pass
                else:
                    logger.debug('    still alive; will try again soon')

        self._instances.clear()
        self._jobs_to_instances.clear()

        logger.debug('Closing event thread')
        self._running = False

        if self._event_thread is not None:
            self._event_thread.join()
            logger.debug('Deleting event thread')
            del self._event_thread

        logger.debug('Closing event queue')

        if self._event_queue is not None:
            self._event_queue.close()
            self._event_queue.join_thread()

        logger.debug('Done')

    @_require_running
    def submit_job(self, job: Job, run_time: datetime) -> int:
        """Submits a `Job` to be executed.

        Args:
            job (Job): A `Job` instance to execute.
            run_time (datetime): When the `Job` is scheduled to run.

        Raises:
            ExecutorNotSetupException: If `setup()` was not run.
            MaxJobInstancesReachedException: If the `Job` already has reached the maximum amount of
                instances allowed to run at one time.

        Returns:
            int: The instance id of the `Job`.
        """
        with self._lock:
            if job.max_instances != -1 and len(self._jobs_to_instances[job.id]) >= job.max_instances:
                raise MaxJobInstancesReachedException(job, self._jobs_to_instances[job.id])

            dynamic_args: dict[str, Any] = {}
            for arg_name, values in job.dynamic_args.items():
                func = values['func']
                dynamic_args[arg_name] = func()

            instance_id = self._scheduler._jobstore.get_new_instance_id()
            process = self.mp_ctx.Process(target=run_job, name=f'Process-{instance_id}',
                                          args=(job, instance_id, run_time, self._event_queue),
                                          kwargs={**job.kwargs, **dynamic_args})

            self._instances[instance_id] = Instance(job,
                                                    instance_id,
                                                    datetime.now(tz=self._scheduler.tz),
                                                    process)
            self._jobs_to_instances[job.id].append(instance_id)

            try:
                logger.info(f'Running job "{job.id}" (scheduled at {run_time})')

                with contextlib.redirect_stdout(None), contextlib.redirect_stderr(None):
                    process.start()
            except Exception:
                self._job_cleanup(job.id, instance_id, terminate=True)
                raise

            return instance_id

    @_require_running
    def terminate_job(self, instance_id: int) -> str:
        """Terminates a `Job` via its `instance_id`.

        Args:
            instance_id (int): The instance id of the desired `Job` to terminate.

        Raises:
            JobInstanceNotFoundException: If there is no instance id matching the one provided.

        Returns:
            str: The `Job`'s id that was terminated.
        """
        with self._lock:
            instance = self._instances.get(instance_id, None)

            if instance is None:
                raise JobInstanceNotFoundException(instance_id)

        job_id = instance.job.id
        self._job_cleanup(job_id, instance_id, terminate=True)

        return job_id

    @_require_running
    def get_job_instance_ids(self, job_id: str) -> list[int]:
        """Returns all instance ids associated with the provided `job_id`.

        Args:
            job_id (str): Identifier for the interested job.

        Returns:
            List[int]: A list containing all instance ids associated with the provided 'job_id'.
        """
        with self._lock:
            return self._jobs_to_instances.get(job_id, [])

    @_require_running
    def get_job(self, instance_id: int) -> Job:
        """Returns the `Job` associated with the provided `instance_id`.

        Args:
            instance_id (int): Instance id used to find the associated `Job`.

        Raises:
            JobInstanceNotFoundException: If the provided `instance_id` was not found.

        Returns:
            Job: A `Job` associated with the provided `instance_id`.
        """
        with self._lock:
            instance = self._instances.get(instance_id, None)
            if instance is None:
                raise JobInstanceNotFoundException(instance_id)

            return instance.job

    @_require_running
    def get_instances(self) -> list[Instance]:
        """Returns all instances and their associated data.

        Returns:
            list[dict]: A list of `Instance`s containing all instances and their associated data.
            Each Instance has the following data:
            {
                'job': <Job Instance>
                'id': <Instance's id>
                'start_time': <Datetime>
                'process': <Process instance>
            }
        """
        with self._lock:
            return list(self._instances.values())

    def is_running(self) -> bool:
        """Returns if the Executor is currently running.

        Returns:
            bool: True when the Executor is running, False otherwise.
        """
        return self._running

    @_require_running
    def _event_read_loop(self) -> None:
        """Processes events as they are put into the `_event_queue`.

        Upon a successful event, the `Job`'s `callback` is called.
        Upon an unsuccessful event, the `Job`'s `error_callback` is called.

        The events are passed along to the scheduler to handle as well.
        """
        logger.debug('waiting for event...')
        if self._event_queue is None:
            raise ExecutorNotSetupException()

        while self._running:
            try:
                event = self._event_queue.get(timeout=0.2)
            except Empty:
                continue

            job = self._instances[event.instance_id].job

            logger.debug(f'got event ({event}) from job {job.id}:{event.instance_id}; '
                         f'event queue size: {self._event_queue.qsize()}')

            if event.success:
                logger.info(f'Job "{job.id}" executed successfully')
                if job.callback:
                    try:
                        job.callback(job.id, event.instance_id, event.retval)
                    except Exception as e:
                        logger.error(f'callback failed!')
                        logger.exception(e)
            else:
                logger.error(f'Job "{job.id}" raised an exception\n{event.formatted_exc}')
                if job.error_callback:
                    try:
                        job.error_callback(job.id, event.instance_id, event.exc)
                    except Exception as e:
                        logger.error(f'error callback failed!')
                        logger.exception(e)

            self._job_cleanup(event.job_id, event.instance_id)
            self._scheduler._dispatch_event(event)

            logger.debug('waiting for event...')

    @_require_running
    def _job_cleanup(self, job_id: str, instance_id: int, terminate: bool = False) -> None:
        """Cleans up a `Job` instance, shutting it down if necessary and requested.

        Args:
            job_id (str): The id for the `Job` to cleanup.
            instance_id (int): The instance id for the `Job` to cleanup.
            terminate (bool, optional): Determines if the `Job` should be terminated.
                Defaults to False.
        """
        with self._lock:
            process = self._instances[instance_id].process

            logger.debug(f'cleaning up instance (job_id: {job_id}, instance_id: {instance_id}, process: {process})')

            if process.is_alive():
                if terminate:
                    logger.debug(f'terminating process')
                    try:
                        process.terminate()
                    except Exception as e:
                        logger.exception(e)
                process.join(timeout=2)

            if not process.is_alive():
                process.close()
                logger.debug('process closed')
            else:
                logger.error('Could not terminate process')

            logger.debug(f'instances: {self._instances}')
            logger.debug(f'mapping: {dict(self._jobs_to_instances)}')

            logger.debug('removing instance')
            del self._instances[instance_id]
            self._jobs_to_instances[job_id].remove(instance_id)

            logger.debug(f'instances: {self._instances}')
            logger.debug(f'mapping: {dict(self._jobs_to_instances)}')

            still_running = [f'{jid}:{str(iids).replace(" ", "")}' for jid, iids in self._jobs_to_instances.items() if iids]
            logger.info(f'Currently running: {", ".join(still_running)}')

            logger.debug('cleaned up')


def run_job(job: Job, instance_id: int, run_time: datetime, event_queue: multiprocessing.Queue, **kwargs: Any) -> None:
    """Executes the provided `Job` by running its `func`.

    Args:
        job (Job): The `Job` to run.
        instance_id (int): The instance id associated with the `job`.
        run_time (datetime): When the `job` is scheduled to run.
        event_queue (multiprocessing.Queue): Used to track the event produced by running the `job`.
    """
    try:
        retval = job.func(job.id, instance_id, *job.args, **job.kwargs, **kwargs)
        logger.debug(f'job {job.id}:{instance_id} returned {retval}')
    except Exception as e:
        logger.debug(f'job {job.id}:{instance_id} raised an exception')
        event = events.JobExecutionEvent(events.EVENT_JOB_EXECUTED, job.id, instance_id, run_time, success=False,
                                         exc=e)
    else:
        logger.debug(f'job {job.id}:{instance_id} executed successfully')
        event = events.JobExecutionEvent(events.EVENT_JOB_EXECUTED, job.id, instance_id, run_time, success=True,
                                         retval=retval)

    logger.debug(f'sending event to queue ({job.id}:{instance_id})')
    for handler in logger.handlers:
        handler.flush()
    event_queue.put(event)
    logger.debug(f'event queue size: {event_queue.qsize()} ({job.id}:{instance_id})')


class ExecutorNotSetupException(Exception):
    """Raised when the `Executor` was not setup yet."""
    def __init__(self) -> None:
        super().__init__('Executor not setup')


class MaxJobInstancesReachedException(Exception):
    """Raised when trying to run a `Job` already running the maximum number of instances.""" 

    def __init__(self, job: Job, instance_ids: list[int]) -> None:
        """Initializes a MaxJobInstancesReachedException.

        Args:
            job (Job): The `Job` which is already running the maximum number of instances.
            instance_ids (List[int]): The list of the `Job`'s instance ids.
        """
        super().__init__(f'Job "{job.id}" has already reached its maximum number of instances '
                         f'({job.max_instances})')
        self.job = job
        self.instance_ids = instance_ids


class JobInstanceNotFoundException(Exception):
    """Raised when the provided instance id was not associated with any `Job`."""

    def __init__(self, instance_id: int) -> None:
        """Initializes a JobInstanceNotFoundException.

        Args:
            instance_id (int): The instance id of which no `Job` was associated with.
        """
        super().__init__(f'Job instance {instance_id} not found')
        self.instance_id = instance_id



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A set of utility methods that are used by the YAP Scheduler."""

from functools import partial
import inspect
import logging
from typing import Any, Callable

logger = logging.getLogger(__name__)


def obj_to_ref(obj: Any) -> str:
    """Returns a string with the path to the given callable.

    Args:
        obj (Callable): An object, expected to be a callable method.

    Raises:
        ValueError: If the given object is a partial, lamda, or nested function.

    Returns:
        str: A `module:name` string path to the callable.
    """
    if isinstance(obj, partial):
        raise ValueError('Cannot create a reference to a partial()')

    name = get_callable_name(obj)
    if '<lambda>' in name:
        raise ValueError('Cannot create a reference to a lambda')
    if '<locals>' in name:
        raise ValueError('Cannot create a reference to a nested function')

    if inspect.ismethod(obj):
        if hasattr(obj, 'im_self') and obj.im_self:
            # bound method
            module = obj.im_self.__module__
        elif hasattr(obj, 'im_class') and obj.im_class:
            # unbound method
            module = obj.im_class.__module__
        else:
            module = obj.__module__
    else:
        module = obj.__module__
    return '%s:%s' % (module, name)


def ref_to_obj(ref: str) -> object:
    """Parses a reference path string to retrieve the object it points to.

    Args:
        ref (str): A string path to an object.

    Raises:
        TypeError: If the reference is not a string.
        ValueError: If it is not a valid reference string (not containg ':').
        LookupError: If the object's module could not be imported or its attribute could not
            be resolved.

    Returns:
        object: The object pointed to by `ref` string.
    """
    if not isinstance(ref, str):
        raise TypeError('References must be strings')
    if ':' not in ref:
        raise ValueError('Invalid reference')

    module_name, rest = ref.split(':', 1)
    try:
        obj = __import__(module_name, fromlist=[rest])
    except ImportError:
        raise LookupError('Error resolving reference %s: could not import module' % ref)

    try:
        for name in rest.split('.'):
            obj = getattr(obj, name)
        return obj
    except Exception:
        raise LookupError('Error resolving reference %s: error looking up object' % ref)


def get_callable_name(func: Callable) -> str:
    """Returns the best available display name for the given callable.

    Args:
        func (Callable): The function to get a name for.

    Raises:
        TypeError: If a name for the function wasn't able to be resolved.

    Returns:
        str: The best available display name for the given callable.
    """
    # the easy case (on Python 3.3+)
    if hasattr(func, '__qualname__'):
        return func.__qualname__

    # class methods, bound and unbound methods
    f_self = getattr(func, '__self__', None) or getattr(func, 'im_self', None)
    if f_self and hasattr(func, '__name__'):
        f_class = f_self if inspect.isclass(f_self) else f_self.__class__
    else:
        f_class = getattr(func, 'im_class', None)

    if f_class and hasattr(func, '__name__'):
        return '%s.%s' % (f_class.__name__, func.__name__)

    # class or class instance
    if hasattr(func, '__call__'):
        # class
        if hasattr(func, '__name__'):
            return func.__name__

        # instance of a class with a __call__ method
        return func.__class__.__name__

    raise TypeError('Unable to determine a name for %r -- maybe it is not a callable?' % func)


def get_class_logger(obj: Any) -> logging.Logger:
    """Returns the logger for the given object.

    Args:
        obj (object): An object to retrive its class logger for.

    Returns:
        logging.Logger: The object's class logger.
    """
    klass = list(filter(lambda m: m[0] == '__class__', inspect.getmembers(obj)))[0][1]
    cls_name = str(klass)[8:-2]
    return logging.getLogger(cls_name)



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the `Job` class, the basic unit of work for the `Scheduler`.

Each `Job`, at minimum, has a unique identifier and a function to execute when it has been
scheduled to run.
"""

from typing import Any, Optional, Callable, Protocol, TypedDict, cast
from datetime import datetime
import inspect
from .triggers import Trigger
from . import util


# pylint: disable=too-few-public-methods
class JobFunction(Protocol):
    """What is called by a `Job` when run.

    Args:
        job_id (str): A unique ID for the job.
        instance_id (int): The instance id for this `Job` process.
        *args: Any additional arguments.
        **kwargs: Any additional keyword arguments.
    """
    def __call__(self,
                 job_id: str,
                 instance_id: int,
                 *args: Any,
                 **kwargs: Any) -> Any: ...

    def __self__(self) -> 'JobFunction':
        pass


class JobCallback(Protocol):
    """What is called after the `Job` executes successfully.

    Args:
        job_id (str): A unique ID for the job.
        instance_id (int): The instance id for this `Job` process.
        retval (Any): The return value from the `JobFunction`.
    """
    def __call__(self,
                 job_id: str,
                 instance_id: int,
                 retval: Any) -> None: ...


class JobErrorCallback(Protocol):
    """What is called after the `Job` fails to execute successfully.

    Args:
        job_id (str): A unique ID for the job.
        instance_id (int): The instance id for this `Job` process.
        exc (Exception): What `Exception` was raised during the `JobFunction`.
    """
    def __call__(self,
                 job_id: str,
                 instance_id: int,
                 exc: Optional[Exception]) -> None: ...


class Job:
    """A job to be executed by the executor.

    Attributes:
        func (JobFunction): The function to be called when the job executes.
        func_ref (str): Path to the `func` function, which is used for `Job` serialization.
        id (str): A unique ID of the job.
        args (Tuple): The arguments to pass to the `func`.
        kwargs (Dict): The keyword arguments to pass to the `func`.
        dynamic_args (Dict[str, Callable]): A dictionary of argument names mapping to functions that
            will be executed upon job execution. the return values of the functions will be passed
            to func as keyword arguments.
        trigger (Trigger): The trigger to used to determine when the job will execute.
        next_run_time (Optional[datetime]): The next run time of the job. If None,
            the next run time will be determined by the trigger. Defaults to None.
        description (Optional[str]): The description of the job. Defaults to None.
        coalesce (bool): If true, it will only run once when several run times are due.
            Defaults to False.
        max_instances (int): The maximum number of concurrently running instances of the
            job. -1 means no limit, 0 means disabled. Defaults to -1.
        active (bool): Determines if the job is active and should be executed.
            Defaults to True.
        callback (Optional[JobCallback]): A function to call if `func` executes successfully.
            It must accept three arguments: `job_id`, `instance_id`, and `retval`. The `retval`
            argument will be the return value of `func`. Defaults to None.
        callback_ref (str): Path to the `callback` function, which is used for `Job` serialization.
        error_callback (Optional[JobErrorCallback]): A function to call if `func` raises an
            exception. It must accept three arguments: `job_id`, `instance_id`, and `exc`.
            The `exc` argument will be the exception instance. Defaults to None.
        error_callback_ref (str): Path to the `error_callback` function, which is used for `Job`
            serialization.
    """

    def __init__(self,
                 func: JobFunction,
                 identifier: str,
                 args: tuple,
                 kwargs: dict[str, Any],
                 dynamic_args: dict[str, Callable],
                 trigger: Trigger,
                 next_run_time: Optional[datetime] = None,
                 description: Optional[str] = None,
                 coalesce: bool = False,
                 max_instances: int = -1,
                 active: bool = True,
                 callback: Optional[JobCallback] = None,
                 error_callback: Optional[JobErrorCallback] = None):
        """Initializes a `Job`.

        Args:
            func (JobFunction): The function to be called when the job executes.
            identifier (str): A unique ID of the job.
            args (Tuple): The arguments to pass to the `func`.
            kwargs (Dict): The keyword arguments to pass to the `func`.
            dynamic_args (Dict[str, Callable]): A dictionary of argument names mapping to functions
                that will be executed upon job execution. The return values of the functions will
                be passed to func as keyword arguments.
            trigger (Trigger): The trigger used to determine when the job will execute.
            next_run_time (Optional[datetime]): The next run time of the job. If None,
                the next run time will be determined by the trigger. Defaults to None.
            description (Optional[str]): The description of the job. Defaults to None.
            coalesce (bool): If true, it will only run once when several run times are
                due.  Defaults to False.
            max_instances (int): The maximum number of concurrently running instances
                of the job. -1 means no limit, 0 means disabled. Defaults to -1.
            active (bool): Determines if the job is active and should be executed.
                Defaults to True.
            callback (Optional[JobCallback]): A function to call if `func` executes
                successfully. It must accept three arguments: `job_id`, `instance_id`, and `retval`.
                The `retval` argument will be the return value of `func`. Defaults to None.
            error_callback (Optional[JobErrorCallback]): A function to call if `func` raises an
                exception. It must accept three arguments: `job_id`, `instance_id`, and `exc`.
                The `exc` argument will be the exception instance. Defaults to None.
        """
        self.func = func
        self.id = identifier
        self.args = args
        self.kwargs = kwargs
        self.trigger = trigger
        self.next_run_time = next_run_time
        self.description = description
        self.coalesce = coalesce
        self.max_instances = max_instances
        self.active = active

        self.func_ref: Optional[str]
        try:
            self.func_ref = util.obj_to_ref(self.func)
        except ValueError:
            self.func_ref = None

        self.callback: JobCallback
        if callback is None:
            self.callback = dummy_callback  # type: ignore
        else:
            self.callback = callback

        self.error_callback: JobErrorCallback
        if error_callback is None:
            self.error_callback = dummy_callback  # type: ignore
        else:
            self.error_callback = error_callback

        self.callback_ref: Optional[str]
        try:
            self.callback_ref = util.obj_to_ref(self.callback)
        except ValueError:
            self.callback_ref = None

        self.error_callback_ref: Optional[str]
        try:
            self.error_callback_ref = util.obj_to_ref(self.error_callback)
        except ValueError:
            self.error_callback_ref = None

        dynamic_arg = TypedDict('dynamic_arg', {'func': Callable[..., Any], 'func_ref': Optional[str]})
        self.dynamic_args: dict[str, dynamic_arg] = {}
        arg_func_ref: Optional[str]
        for arg_name, arg_func in dynamic_args.items():
            try:
                arg_func_ref = util.obj_to_ref(arg_func)
            except ValueError:
                arg_func_ref = None

            self.dynamic_args[arg_name] = {'func': arg_func, 'func_ref': arg_func_ref}

    def _modify(self, **changes: dict[str, Any]) -> None:
        pass

    def _get_run_times(self, latest: datetime) -> list[datetime]:
        """Returns a list of all the upcoming datetimes when the job is scheduled to run.

        Args:
            latest (datetime): Exclusionary boundary for upcoming jobs - serves as an 'up to' time.

        Returns:
            List[datetime]: A list of all the upcoming datetimes when th job is scheduled to run.
        """
        run_times: list[datetime] = []
        next_run_time = self.next_run_time
        while next_run_time and next_run_time < latest:
            run_times.append(next_run_time)
            next_run_time = self.trigger.get_next_fire_time(next_run_time, latest)
        return run_times

    def get_next_run_time_ts(self) -> Optional[float]:
        """Returns the timestamp for the next scheduled run time of the job."""
        return self.next_run_time.timestamp() if self.next_run_time is not None else None

    def __getstate__(self) -> dict[str, Any]:
        """Returns the current state of the job, used for pickling.

        Raises:
            Exception: If the `func_ref`, `callback_ref`, `error_callback_ref`, or
                `func_ref` of any `dynamic_args` is None.

        Returns:
            Dict[str, Any]: Contains a mapping of `Job` attributes to their values, except
                the function reference paths.
        """
        if self.func_ref is None:
            raise Exception(f'Job function ({self.func}) not serializable')

        if self.callback_ref is None:
            raise Exception(f'Job callback function ({self.callback}) not serializable')

        if self.error_callback_ref is None:
            raise Exception(f'Job error callback function ({self.error_callback}) not serializable')

        dynamic_args: dict[str, str] = {}
        for arg_name, values in self.dynamic_args.items():
            func = values['func']
            func_ref = values['func_ref']
            if func_ref is None:
                raise Exception(f'Job dynamic arg function ({func}) not serializable')
            dynamic_args[arg_name] = func_ref

        if inspect.ismethod(self.func) and not inspect.isclass(self.func.__self__):
            args = (self.func.__self__,) + tuple(self.args)
        else:
            args = self.args

        return {
            'id': self.id,
            'func': self.func_ref,
            'args': args,
            'kwargs': self.kwargs,
            'dynamic_args': dynamic_args,
            'trigger': self.trigger,
            'next_run_time': self.next_run_time,
            'description': self.description,
            'coalesce': self.coalesce,
            'max_instances': self.max_instances,
            'active': self.active,
            'callback': self.callback_ref,
            'error_callback': self.error_callback_ref,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Explicitly sets all `Job` attributes, used when unpickling.

        Args:
            state (Dict): Mapping of `Job` attributes to their respective values.
        """
        self.id = state['id']

        self.func_ref = state['func']
        self.func = util.ref_to_obj(self.func_ref)  # type: ignore

        self.args = state['args']
        self.kwargs = state['kwargs']
        self.trigger = state['trigger']
        self.next_run_time = state['next_run_time']
        self.description = state['description']
        self.coalesce = state['coalesce']
        self.max_instances = state['max_instances']
        self.active = state['active']

        self.callback_ref = state['callback']
        self.callback = util.ref_to_obj(self.callback_ref)  # type: ignore

        self.error_callback_ref = state['error_callback']
        self.error_callback = util.ref_to_obj(self.error_callback_ref)  # type: ignore

        self.dynamic_args = {}
        for arg_name, func_ref in state['dynamic_args'].items():
            func = cast(Callable[..., Any], util.ref_to_obj(func_ref))
            func_ref = cast(Optional[str], func_ref)
            self.dynamic_args[arg_name] = {'func': func, 'func_ref': func_ref}

    def __eq__(self, other: object) -> bool:
        """`Job`s are equal when they share identical `id`s.

        Args:
            other (Any): An object to be compared against.

        Returns:
            bool: True if `other` is a `Job` and shares an identical `id`, False otherwise.
        """
        if isinstance(other, Job):
            return self.id == other.id
        raise NotImplemented

    def __repr__(self) -> str:
        return f'<Job (id={self.id})>'

    def __str__(self) -> str:
        status = f'next run at: {self.next_run_time}' if self.next_run_time is not None else 'INACTIVE'
        return f'{self.id} (trigger: {self.trigger}, {status})'


def dummy_callback(x, y, z) -> None:  # type: ignore
    pass
